"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDataForFingerprintsByQuery = exports.getDataForFingerprintsByTimestampAtLeast = exports.getDataForFingerprint = exports.unpack = void 0;
const axios_1 = __importDefault(require("axios"));
/*
Example "queryResponse":
{
        "_": {},
        "response": {
            "version": "1.2.0",
            "seq": 6,
            "encoding": "rle",
            "columns": [
                [
                    "head(callstack)",
                    "callstack"
                ],
                [
                    "head(classifiers)",
                    ""
                ],
                [
                    "head(fingerprint;first_seen)",
                    "unix_timestamp"
                ]
            ],
            "factors_desc": [
                {
                    "name": "fingerprint",
                    "format": "sha256",
                    "type": "dictionary"
                }
            ],
            "columns_desc": [
                {
                    "name": "callstack",
                    "format": "callstack",
                    "type": "dictionary",
                    "op": "head"
                },
                {
                    "name": "classifiers",
                    "format": "",
                    "type": "dictionary",
                    "op": "head"
                },
                {
                    "name": "fingerprint;first_seen",
                    "format": "unix_timestamp",
                    "type": "uint64",
                    "op": "head"
                }
            ],
            "values": [
                [
                    "16a75b86ba7ecc96a0c40167356ec77b2c09aed48e504f8b3b268cbd147379bb",
                    [
                        [
                            "{ \"frame\": [ \"foo\",  \"bar\" ] }"
                        ],
                        [
                            "segv stop null memory.read"
                        ],
                        [
                            "1595596927"
                        ]
                    ],
                    378
                ]
            ]
        }
    }
*/
function unpack(queryResponse) {
    const result = {}; // for unpacking
    queryResponse.values.forEach((value) => {
        const fp = value[0];
        const colValues = value[1];
        if (fp && Array.isArray(colValues) && !result[fp]) {
            result[fp] = {};
            result[fp].count = value.length > 2 ? value[2] : 0;
            queryResponse.columns.forEach((column, index) => {
                const label = column[0];
                result[fp][label] = colValues[index];
            });
        }
    });
    return result;
}
exports.unpack = unpack;
async function getDataForFingerprint(fp, url) {
    const query = {
        limit: 1,
        order: [{ name: ';count', ordering: 'descending' }],
        group: ['fingerprint'],
        fold: {
            callstack: [['head']],
            classifiers: [['head']],
            'fingerprint;first_seen': [['head']],
            timestamp: [['range']],
            'fingerprint;issues;state': [['head']],
        },
        filter: [
            {
                timestamp: [['at-least', 1]],
                fingerprint: [['equal', fp]],
            },
        ],
    };
    const results = await axios_1.default.post(url, query);
    return unpack(results.data.response);
}
exports.getDataForFingerprint = getDataForFingerprint;
async function getDataForFingerprintsByTimestampAtLeast(timestamp, url, limit) {
    const query = {
        limit: limit + 1,
        order: [{ name: ';count', ordering: 'descending' }],
        group: ['fingerprint'],
        fold: {
            callstack: [['head']],
            classifiers: [['head']],
            'fingerprint;first_seen': [['head']],
            timestamp: [['range']],
            'fingerprint;issues;state': [['head']],
        },
        filter: [
            {
                timestamp: [['at-least', timestamp]],
            },
        ],
    };
    const results = await axios_1.default.post(url, query);
    return unpack(results.data.response);
}
exports.getDataForFingerprintsByTimestampAtLeast = getDataForFingerprintsByTimestampAtLeast;
async function getDataForFingerprintsByQuery(url, filter, limit) {
    const query = {
        limit: limit + 1,
        order: [{ name: ';count', ordering: 'descending' }],
        group: ['fingerprint'],
        fold: {
            callstack: [['head']],
            classifiers: [['head']],
            'fingerprint;first_seen': [['head']],
            timestamp: [['range']],
            'fingerprint;issues;state': [['head']],
        },
        filter,
    };
    const results = await axios_1.default.post(url, query);
    return unpack(results.data.response);
}
exports.getDataForFingerprintsByQuery = getDataForFingerprintsByQuery;
//# sourceMappingURL=candidates.js.map